import pytest
import json


@pytest.mark.asyncio
async def test_count_nodes(mcp_client):
    result = await mcp_client.call_tool("count_nodes")

    assert len(result) == 1
    result_text = result[0]["text"]
    node_count = int(result_text.strip())
    assert node_count == 302


@pytest.mark.asyncio
async def test_list_tools(mcp_client):
    """Test that all expected tools are listed."""

    # Get list of available tools
    tools = await mcp_client.list_tools()
    tool_names = [tool["name"] for tool in tools]

    # Expected tools (based on server.py imports)
    expected_tools = [
        # Basic tools
        "count_nodes",
        "get_node_properties_keys",
        "get_relationship_properties_keys",
        "get_node_labels",
        # Centrality algorithms
        "article_rank",
        "articulation_points",
        "betweenness_centrality",
        "bridges",
        "CELF",
        "closeness_centrality",
        "degree_centrality",
        "eigenvector_centrality",
        "pagerank",
        "harmonic_centrality",
        "HITS",
        # Community algorithms
        "conductance",
        "HDBSCAN",
        "k_core_decomposition",
        "k_1_coloring",
        "k_means_clustering",
        "label_propagation",
        "leiden",
        "local_clustering_coefficient",
        "louvain",
        "modularity_metric",
        "modularity_optimization",
        "strongly_connected_components",
        "triangle_count",
        "weakly_connected_components",
        "approximate_maximum_k_cut",
        "speaker_listener_label_propagation",
        # Path algorithms
        "find_shortest_path",
        "delta_stepping_shortest_path",
        "dijkstra_single_source_shortest_path",
        "a_star_shortest_path",
        "yens_shortest_paths",
        "minimum_weight_spanning_tree",
        "minimum_directed_steiner_tree",
        "prize_collecting_steiner_tree",
        "all_pairs_shortest_paths",
        "random_walk",
        "breadth_first_search",
        "depth_first_search",
        "bellman_ford_single_source_shortest_path",
        "longest_path",
        # similarity
        "node_similarity",
        "k_nearest_neighbors",
    ]

    # Check that we have the expected tools
    assert len(tool_names) == len(expected_tools)
    for expected_tool in expected_tools:
        assert expected_tool in tool_names, (
            f"Expected tool '{expected_tool}' not found in tool list"
        )


@pytest.mark.asyncio
async def test_get_node_properties_keys(mcp_client):
    result = await mcp_client.call_tool("get_node_properties_keys")

    assert len(result) == 1
    result_text = result[0]["text"]
    properties_keys = json.loads(result_text)

    assert properties_keys == [
        "zone",
        "rail",
        "latitude",
        "name",
        "total_lines",
        "id",
        "display_name",
        "longitude",
    ]


@pytest.mark.asyncio
async def test_get_relationship_properties_keys(mcp_client):
    result = await mcp_client.call_tool("get_relationship_properties_keys")

    assert len(result) == 1
    result_text = result[0]["text"]
    properties_keys = json.loads(result_text)

    assert properties_keys == ["distance", "line", "time"]


@pytest.mark.asyncio
async def test_get_node_labels(mcp_client):
    result = await mcp_client.call_tool("get_node_labels")

    assert len(result) == 1
    result_text = result[0]["text"]
    properties_keys = json.loads(result_text)

    assert properties_keys == ["UndergroundStation"]



from mcp import types

community_tool_definitions = [
    types.Tool(
        name="conductance",
        description="""Calculate the conductance metric for all communities""",
        inputSchema={
            "type": "object",
            "properties": {
                "communityProperty": {
                    "type": "string",
                    "description": "The node property that holds the community ID as an integer for each node. "
                    "Note that only non-negative community IDs are considered valid and will have their conductance computed.",
                },
                "relationshipWeightProperty": {
                    "type": "string",
                    "description": "The relationship property that holds the weight of the relationships. "
                    "If not provided, all relationships are considered to have a weight of 1.",
                },
                "nodeLabels": {
                    "type": "array",
                    "items": {"type": "string"},
                    "description": "The node labels used to project and run conductance on. Nodes with different node labels will be ignored. Do not specify to run for all nodes",
                },
            },
            "required": ["communityProperty"],
        },
    ),
    types.Tool(
        name="HDBSCAN",
        description="Perform HDBSCAN clustering on the graph. HDBSCAN, which stands for Hierarchical Density-Based Spatial Clustering of Applications with Noise, "
        "is a clustering algorithm used to identify clusters of similar data points within a dataset. "
        "It builds upon the DBSCAN algorithm but adds a hierarchical structure, making it more robust to varying densities within the data. "
        "Unlike DBSCAN, HDBSCAN does not require tuning a specific density parameter; "
        "instead, it runs DBSCAN over a range of parameters, creating a hierarchy of clusters. "
        "This hierarchical approach allows HDBSCAN to find clusters of varying densities and to be more adaptable to real-world data.HDBSCAN is known for its ease of use, "
        "noise tolerance, and ability to handle data with varying densities, making it a versatile tool for clustering tasks, "
        "especially when dealing with complex, high-dimensional datasets.",
        inputSchema={
            "type": "object",
            "properties": {
                "nodeProperty": {
                    "type": "string",
                    "description": "A node property corresponding to an array of floats used by HDBSCAN to compute clusters",
                },
                "nodeIdentifierProperty": {
                    "type": "string",
                    "description": "Property name to use for identifying nodes (e.g., 'name', 'Name', 'title'). Use get_node_properties_keys to find available properties.",
                },
                "minClusterSize": {
                    "type": "integer",
                    "description": "The minimum number of nodes that a cluster should contain.",
                },
                "samples": {
                    "type": "integer",
                    "description": "The number of neighbours to be considered when computing the core distances of a node.",
                },
                "leafSize": {
                    "type": "integer",
                    "description": "The number of leaf nodes of the supporting tree data structure.",
                },
                "nodeLabels": {
                    "type": "array",
                    "items": {"type": "string"},
                    "description": "The node labels used to project and run HDBSCAN on. Nodes with different node labels will be ignored. Do not specify to run for all nodes",
                },
            },
            "required": ["nodeProperty"],
        },
    ),
    types.Tool(
        name="k_core_decomposition",
        description="The K-core decomposition constitutes a process of separates the nodes in a graph into groups based on the degree sequence and topology of the graph. "
        "The term i-core refers to a maximal subgraph of the original graph such that each node in this subgraph has degree at least i. "
        "The maximality ensures that it is not possible to find another subgraph with more nodes where this degree property holds. "
        "The nodes in the subgraph denoted by i-core also belong to the subgraph denoted by j-core for any j<i. The converse however is not true.  "
        "Each node u is associated with a core value which denotes the largest value i such that u belongs to the i-core. "
        "The largest core value is called the degeneracy of the graph.Standard algorithms for K-Core Decomposition iteratively remove the node of lowest degree until the graph becomes empty. "
        "When a node is removed from the graph, all of its relationships are removed, and the degree of its neighbors is reduced by one. "
        "With this approach, the different core groups are discovered one-by-one.",
        inputSchema={
            "type": "object",
            "properties": {
                "nodeIdentifierProperty": {
                    "type": "string",
                    "description": "Property name to use for identifying nodes (e.g., 'name', 'Name', 'title'). Use get_node_properties_keys to find available properties.",
                },
                "nodeLabels": {
                    "type": "array",
                    "items": {"type": "string"},
                    "description": "The node labels used to project and run K Core decomposition on. Nodes with different node labels will be ignored. Do not specify to run for all nodes",
                },
            },
            "required": [],
        },
    ),
    types.Tool(
        name="k_1_coloring",
        description="The K-1 Coloring algorithm assigns a color to every node in the graph, "
        "trying to optimize for two objectives: "
        "1. To make sure that every neighbor of a given node has a different color than the node itself. "
        "2. To use as few colors as possible. "
        "Note that the graph coloring problem is proven to be NP-complete, which makes it intractable on anything but trivial graph sizes. "
        "For that reason the implemented algorithm is a greedy algorithm. "
        "Thus it is neither guaranteed that the result is an optimal solution, using as few colors as theoretically possible, "
        "nor does it always produce a correct result where no two neighboring nodes have different colors. "
        "However the precision of the latter can be controlled by the number of iterations this algorithm runs.",
        inputSchema={
            "type": "object",
            "properties": {
                "nodeIdentifierProperty": {
                    "type": "string",
                    "description": "Property name to use for identifying nodes (e.g., 'name', 'Name', 'title'). Use get_node_properties_keys to find available properties.",
                },
                "nodeLabels": {
                    "type": "array",
                    "items": {"type": "string"},
                    "description": "The node labels used to project and run K-1 Coloring on. Nodes with different node labels will be ignored. Do not specify to run for all nodes",
                },
                "maxIterations": {
                    "type": "integer",
                    "description": "The maximum number of iterations to run the coloring algorithm.",
                },
                "minCommunitySize": {
                    "type": "integer",
                    "description": "Only nodes inside communities larger or equal the given value are returned.",
                },
            },
            "required": [],
        },
    ),
    types.Tool(
        name="k_means_clustering",
        description="K-Means clustering is an unsupervised learning algorithm that is used to solve clustering problems. "
        "It follows a simple procedure of classifying a given data set into a number of clusters, defined by the parameter k. "
        "The Neo4j GDS Library conducts clustering based on node properties, with a float array node property being passed as input via the nodeProperty parameter. "
        "Nodes in the graph are then positioned as points in a d-dimensional space (where d is the length of the array property). "
        "The algorithm then begins by selecting k initial cluster centroids, which are d-dimensional arrays (see section below for more details). "
        "The centroids act as representatives for a cluster. "
        "Then, all nodes in the graph calculate their Euclidean distance from each of the cluster centroids and are assigned to the cluster of minimum distance from them. "
        "After these assignments, each cluster takes the mean of all nodes (as points) assigned to it to form its new representative centroid (as a d-dimensional array). "
        "The process repeats with the new centroids until results stabilize, i.e., only a few nodes change clusters per iteration or the number of maximum iterations is reached. "
        "Note that the K-Means implementation ignores relationships as it is only focused on node properties.",
        inputSchema={
            "type": "object",
            "properties": {
                "nodeProperty": {
                    "type": "string",
                    "description": "A node property corresponding to an array of floats used by K-Means to cluster nodes into communities.",
                },
                "nodeIdentifierProperty": {
                    "type": "string",
                    "description": "Property name to use for identifying nodes (e.g., 'name', 'Name', 'title'). Use get_node_properties_keys to find available properties.",
                },
                "nodeLabels": {
                    "type": "array",
                    "items": {"type": "string"},
                    "description": "The node labels used to project and run K-Means on. Nodes with different node labels will be ignored. Do not specify to run for all nodes",
                },
                "k": {
                    "type": "integer",
                    "description": "The number of clusters to create.",
                },
                "maxIterations": {
                    "type": "integer",
                    "description": "The maximum number of iterations the algorithm will run.",
                },
                "deltaThreshold": {
                    "type": "number",
                    "description": "Value as a percentage to determine when to stop early. If fewer than 'deltaThreshold * |nodes|' nodes change their cluster , the algorithm stops. Value must be between 0 (exclusive) and 1 (inclusive).",
                },
                "numberOfRestarts": {
                    "type": "integer",
                    "description": "Number of times to execute K-Means with different initial centers. The communities returned are those minimizing the average node-center distances.",
                },
                "initialSampler": {
                    "type": "string",
                    "enum": ["uniform", "kmeans++"],
                    "description": "The method used to sample the first k centroids. 'uniform' and 'kmeans++', both case-insensitive, are valid inputs.",
                },
                "seedCentroids": {
                    "type": "array",
                    "items": {"type": "array", "items": {"type": "number"}},
                    "description": "Parameter to explicitly give the initial centroids. It cannot be enabled together with a non-default value of the numberOfRestarts parameter.",
                },
                "computeSilhouette": {
                    "type": "boolean",
                    "description": "If set to true, the silhouette scores are computed once the clustering has been determined. Silhouette is a metric on how well the nodes have been clustered.",
                },
            },
            "required": ["nodeProperty"],
        },
    ),
    types.Tool(
        name="label_propagation",
        description="The Label Propagation algorithm (LPA) is a fast algorithm for finding communities in a graph. "
        "It detects these communities using network structure alone as its guide, and doesn't require a pre-defined objective function or prior information about the communities. "
        "LPA works by propagating labels throughout the network and forming communities based on this process of label propagation.",
        inputSchema={
            "type": "object",
            "properties": {
                "maxIterations": {
                    "type": "integer",
                    "description": "The maximum number of iterations to run.",
                },
                "nodeWeightProperty": {
                    "type": "string",
                    "description": "The name of a node property that contains node weights.",
                },
                "relationshipWeightProperty": {
                    "type": "string",
                    "description": "Name of the relationship property to use as weights. If unspecified, the algorithm runs unweighted.",
                },
                "seedProperty": {
                    "type": "string",
                    "description": "The name of a node property that defines an initial numeric label.",
                },
                "consecutiveIds": {
                    "type": "boolean",
                    "description": "Flag to decide whether component identifiers are mapped into a consecutive id space (requires additional memory).",
                },
                "minCommunitySize": {
                    "type": "integer",
                    "description": "Only nodes inside communities larger or equal the given value are returned.",
                },
                "nodeIdentifierProperty": {
                    "type": "string",
                    "description": "The name of a node property to use as node identifier in the result. If provided, the result will include a 'nodeName' column with values from this property.",
                },
                "nodeLabels": {
                    "type": "array",
                    "items": {"type": "string"},
                    "description": "The node labels used to project and run Label Propagation on. Nodes with different node labels will be ignored. Do not specify to run for all nodes",
                },
            },
        },
    ),
    types.Tool(
        name="leiden",
        description="The Leiden algorithm is an algorithm for detecting communities in large networks. "
        "The algorithm separates nodes into disjoint communities so as to maximize a modularity score for each community. "
        "Modularity quantifies the quality of an assignment of nodes to communities, that is how densely connected nodes in a community are, compared to how connected they would be in a random network. "
        "The Leiden algorithm is a hierarchical clustering algorithm, that recursively merges communities into single nodes by greedily optimizing the modularity and the process repeats in the condensed graph. "
        "It modifies the Louvain algorithm to address some of its shortcomings, namely the case where some of the communities found by Louvain are not well-connected. "
        "This is achieved by periodically randomly breaking down communities into smaller well-connected ones.",
        inputSchema={
            "type": "object",
            "properties": {
                "maxLevels": {
                    "type": "integer",
                    "description": "The maximum number of levels in which the graph is clustered and then condensed.",
                },
                "gamma": {
                    "type": "number",
                    "description": "Resolution parameter used when computing the modularity. Internally the value is divided by the number of relationships for an unweighted graph, or the sum of weights of all relationships otherwise.",
                },
                "theta": {
                    "type": "number",
                    "description": "Controls the randomness while breaking a community into smaller ones.",
                },
                "tolerance": {
                    "type": "number",
                    "description": "Minimum change in modularity between iterations. If the modularity changes less than the tolerance value, the result is considered stable and the algorithm returns.",
                },
                "includeIntermediateCommunities": {
                    "type": "boolean",
                    "description": "Indicates whether to write intermediate communities. If set to false, only the final community is persisted.",
                },
                "seedProperty": {
                    "type": "string",
                    "description": "Used to set the initial community for a node. The property value needs to be a non-negative number.",
                },
                "minCommunitySize": {
                    "type": "integer",
                    "description": "Only nodes inside communities larger or equal the given value are returned.",
                },
                "nodeIdentifierProperty": {
                    "type": "string",
                    "description": "The name of a node property to use as node identifier in the result. If provided, the result will include a 'nodeName' column with values from this property.",
                },
                "nodeLabels": {
                    "type": "array",
                    "items": {"type": "string"},
                    "description": "The node labels used to project and run Leiden on. Nodes with different node labels will be ignored. Do not specify to run for all nodes",
                },
            },
        },
    ),
    types.Tool(
        name="local_clustering_coefficient",
        description="The Local Clustering Coefficient algorithm computes the local clustering coefficient for each node in the graph. "
        "The local clustering coefficient Cn of a node n describes the likelihood that the neighbours of n are also connected. "
        "To compute Cn we use the number of triangles a node is a part of Tn, and the degree of the node dn. "
        "The formula to compute the local clustering coefficient is as follows: Cn = 2 * Tn / (dn * (dn - 1))"
        "As we can see the triangle count is required to compute the local clustering coefficient. "
        "To do this the Triangle Count algorithm is utilised. "
        "Additionally, the algorithm can compute the average clustering coefficient for the whole graph. "
        "This is the normalised sum over all the local clustering coefficients.",
        inputSchema={
            "type": "object",
            "properties": {
                "triangleCountProperty": {
                    "type": "string",
                    "description": "Node property that contains pre-computed triangle count.",
                },
                "nodeIdentifierProperty": {
                    "type": "string",
                    "description": "The name of a node property to use as node identifier in the result. If provided, the result will include a 'nodeName' column with values from this property.",
                },
                "nodes": {
                    "type": "array",
                    "items": {"type": "string"},
                    "description": "Optional list of node names to filter results. Only nodes whose names (based on nodeIdentifierProperty) contain any of these values will be included in the results. Requires nodeIdentifierProperty to be specified.",
                },
                "nodeLabels": {
                    "type": "array",
                    "items": {"type": "string"},
                    "description": "The node labels used to project and run Local Clustering Coefficient on. Nodes with different node labels will be ignored. Do not specify to run for all nodes",
                },
            },
        },
    ),
    types.Tool(
        name="louvain",
        description="The Louvain method is an algorithm to detect communities in large networks. "
        "It maximizes a modularity score for each community, where the modularity quantifies the quality of an assignment of nodes to communities. "
        "This means evaluating how much more densely connected the nodes within a community are, compared to how connected they would be in a random network. "
        "The Louvain algorithm is a hierarchical clustering algorithm, that recursively merges communities into a single node and executes the modularity clustering on the condensed graphs.",
        inputSchema={
            "type": "object",
            "properties": {
                "relationshipWeightProperty": {
                    "type": "string",
                    "description": "Name of the relationship property to use as weights. If unspecified, the algorithm runs unweighted.",
                },
                "seedProperty": {
                    "type": "string",
                    "description": "Used to set the initial community for a node. The property value needs to be a non-negative number.",
                },
                "maxLevels": {
                    "type": "integer",
                    "description": "The maximum number of levels in which the graph is clustered and then condensed.",
                },
                "maxIterations": {
                    "type": "integer",
                    "description": "The maximum number of iterations that the modularity optimization will run for each level.",
                },
                "tolerance": {
                    "type": "number",
                    "description": "Minimum change in modularity between iterations. If the modularity changes less than the tolerance value, the result is considered stable and the algorithm returns.",
                },
                "includeIntermediateCommunities": {
                    "type": "boolean",
                    "description": "Indicates whether to write intermediate communities. If set to false, only the final community is persisted.",
                },
                "consecutiveIds": {
                    "type": "boolean",
                    "description": "Flag to decide whether component identifiers are mapped into a consecutive id space (requires additional memory). Cannot be used in combination with the includeIntermediateCommunities flag.",
                },
                "minCommunitySize": {
                    "type": "integer",
                    "description": "Only nodes inside communities larger or equal the given value are returned.",
                },
                "nodeIdentifierProperty": {
                    "type": "string",
                    "description": "The name of a node property to use as node identifier in the result. If provided, the result will include a 'nodeName' column with values from this property.",
                },
                "nodeLabels": {
                    "type": "array",
                    "items": {"type": "string"},
                    "description": "The node labels used to project and run Louvain on. Nodes with different node labels will be ignored. Do not specify to run for all nodes",
                },
            },
        },
    ),
    types.Tool(
        name="modularity_metric",
        description="Modularity is a metric that allows you to evaluate the quality of a community detection. "
        "Relationships of nodes in a community C connect to nodes either within C or outside C. "
        "Graphs with high modularity have dense connections between the nodes within communities but sparse connections between nodes in different communities.",
        inputSchema={
            "type": "object",
            "properties": {
                "communityProperty": {
                    "type": "string",
                    "description": "The node property that holds the community ID as an integer for each node. Note that only non-negative community IDs are considered valid and will have their modularity score computed.",
                },
                "relationshipWeightProperty": {
                    "type": "string",
                    "description": "Name of the relationship property to use as weights. If unspecified, the algorithm runs unweighted.",
                },
                "nodeLabels": {
                    "type": "array",
                    "items": {"type": "string"},
                    "description": "The node labels used to project and run Modularity on. Nodes with different node labels will be ignored. Do not specify to run for all nodes",
                },
            },
            "required": ["communityProperty"],
        },
    ),
    types.Tool(
        name="modularity_optimization",
        description="The Modularity Optimization algorithm tries to detect communities in the graph based on their modularity. "
        "Modularity is a measure of the structure of a graph, measuring the density of connections within a module or community. "
        "Graphs with a high modularity score will have many connections within a community but only few pointing outwards to other communities. "
        "The algorithm will explore for every node if its modularity score might increase if it changes its community to one of its neighboring nodes.",
        inputSchema={
            "type": "object",
            "properties": {
                "maxIterations": {
                    "type": "integer",
                    "description": "The maximum number of iterations to run.",
                },
                "tolerance": {
                    "type": "number",
                    "description": "Minimum change in modularity between iterations. If the modularity changes less than the tolerance value, the result is considered stable and the algorithm returns.",
                },
                "seedProperty": {
                    "type": "string",
                    "description": "Used to define initial set of labels.",
                },
                "consecutiveIds": {
                    "type": "boolean",
                    "description": "Flag to decide whether component identifiers are mapped into a consecutive id space (requires additional memory).",
                },
                "relationshipWeightProperty": {
                    "type": "string",
                    "description": "Name of the relationship property to use as weights. If unspecified, the algorithm runs unweighted.",
                },
                "minCommunitySize": {
                    "type": "integer",
                    "description": "Only nodes inside communities larger or equal the given value are returned.",
                },
                "nodeIdentifierProperty": {
                    "type": "string",
                    "description": "The name of a node property to use as node identifier in the result. If provided, the result will include a 'nodeName' column with values from this property.",
                },
                "nodeLabels": {
                    "type": "array",
                    "items": {"type": "string"},
                    "description": "The node labels used to project and run modularity optimization on. Nodes with different node labels will be ignored. Do not specify to run for all nodes",
                },
            },
        },
    ),
    types.Tool(
        name="strongly_connected_components",
        description="The Strongly Connected Components (SCC) algorithm finds maximal sets of connected nodes in a directed graph. "
        "A set is considered a strongly connected component if there is a directed path between each pair of nodes within the set. "
        "It is often used early in a graph analysis process to help us get an idea of how our graph is structured.",
        inputSchema={
            "type": "object",
            "properties": {
                "consecutiveIds": {
                    "type": "boolean",
                    "description": "Flag to decide whether component identifiers are mapped into a consecutive id space (requires additional memory).",
                },
                "nodeIdentifierProperty": {
                    "type": "string",
                    "description": "The name of a node property to use as node identifier in the result. If provided, the result will include a 'nodeName' column with values from this property.",
                },
                "nodeLabels": {
                    "type": "array",
                    "items": {"type": "string"},
                    "description": "The node labels used to project and run SCC on. Nodes with different node labels will be ignored. Do not specify to run for all nodes",
                },
            },
        },
    ),
    types.Tool(
        name="triangle_count",
        description="The Triangle Count algorithm counts the number of triangles for each node in the graph. "
        "A triangle is a set of three nodes where each node has a relationship to the other two. "
        "In graph theory terminology, this is sometimes referred to as a 3-clique. "
        "The Triangle Count algorithm in the GDS library only finds triangles in undirected graphs. "
        "Triangle counting has gained popularity in social network analysis, where it is used to detect communities and measure the cohesiveness of those communities. "
        "It can also be used to determine the stability of a graph, and is often used as part of the computation of network indices, such as clustering coefficients. "
        "The Triangle Count algorithm is also used to compute the Local Clustering Coefficient.",
        inputSchema={
            "type": "object",
            "properties": {
                "maxDegree": {
                    "type": "integer",
                    "description": "If a node has a degree higher than this it will not be considered by the algorithm. The triangle count for these nodes will be -1.",
                },
                "nodeIdentifierProperty": {
                    "type": "string",
                    "description": "The name of a node property to use as node identifier in the result. If provided, the result will include a 'nodeName' column with values from this property.",
                },
                "nodes": {
                    "type": "array",
                    "items": {"type": "string"},
                    "description": "Optional list of node names to filter results. Only nodes whose names (based on nodeIdentifierProperty) contain any of these values will be included in the results. Requires nodeIdentifierProperty to be specified.",
                },
                "nodeLabels": {
                    "type": "array",
                    "items": {"type": "string"},
                    "description": "The node labels used to project and run Triangle Count on. Nodes with different node labels will be ignored. Do not specify to run for all nodes",
                },
            },
        },
    ),
    types.Tool(
        name="weakly_connected_components",
        description="The Weakly Connected Components (WCC) algorithm finds sets of connected nodes in directed and undirected graphs. "
        "Two nodes are connected, if there exists a path between them. The set of all nodes that are connected with each other form a component. "
        "In contrast to Strongly Connected Components (SCC), the direction of relationships on the path between two nodes is not considered. "
        "For example, in a directed graph (a)→(b), a and b will be in the same component, even if there is no directed relationship (b)→(a). "
        "WCC is often used early in an analysis to understand the structure of a graph. "
        "Using WCC to understand the graph structure enables running other algorithms independently on an identified cluster.",
        inputSchema={
            "type": "object",
            "properties": {
                "relationshipWeightProperty": {
                    "type": "string",
                    "description": "Name of the relationship property to use as weights. If unspecified, the algorithm runs unweighted.",
                },
                "seedProperty": {
                    "type": "string",
                    "description": "Used to set the initial component for a node. The property value needs to be a number.",
                },
                "threshold": {
                    "type": "number",
                    "description": "The value of the weight above which the relationship is considered in the computation.",
                },
                "consecutiveIds": {
                    "type": "boolean",
                    "description": "Flag to decide whether component identifiers are mapped into a consecutive id space (requires additional memory).",
                },
                "minComponentSize": {
                    "type": "integer",
                    "description": "Only nodes inside communities larger or equal the given value are returned.",
                },
                "nodeIdentifierProperty": {
                    "type": "string",
                    "description": "The name of a node property to use as node identifier in the result. If provided, the result will include a 'nodeName' column with values from this property.",
                },
                "nodeLabels": {
                    "type": "array",
                    "items": {"type": "string"},
                    "description": "The node labels used to project and run WCC on. Nodes with different node labels will be ignored. Do not specify to run for all nodes",
                },
            },
        },
    ),
    types.Tool(
        name="approximate_maximum_k_cut",
        description="A k-cut of a graph is an assignment of its nodes into k disjoint communities. "
        "So for example a 2-cut of a graph with nodes a,b,c,d could be the communities {a,b,c} and {d}. "
        "A Maximum k-cut is a k-cut such that the total weight of relationships between nodes from different communities in the k-cut is maximized. "
        "That is, a k-cut that maximizes the sum of weights of relationships whose source and target nodes are assigned to different communities in the k-cut. "
        "Suppose in the simple a,b,c,d node set example above we only had one relationship b → c, and it was of weight 1.0. "
        "The 2-cut we outlined above would then not be a maximum 2-cut (with a cut cost of 0.0), whereas for example the 2-cut with communities {a,b} and {c,d} would be one (with a cut cost of 1.0). "
        "n practice, finding the best cut is not feasible for larger graphs and only an approximation can be computed in reasonable time. "
        "The approximate heuristic algorithm implemented in GDS is a parallelized GRASP style algorithm optionally enhanced (via config) with variable neighborhood search (VNS)",
        inputSchema={
            "type": "object",
            "properties": {
                "k": {
                    "type": "integer",
                    "description": "The number of disjoint communities the nodes will be divided into.",
                },
                "iterations": {
                    "type": "integer",
                    "description": "The number of iterations the algorithm will run before returning the best solution among all the iterations.",
                },
                "vnsMaxNeighborhoodOrder": {
                    "type": "integer",
                    "description": "The maximum number of nodes VNS will swap when perturbing solutions.",
                },
                "relationshipWeightProperty": {
                    "type": "string",
                    "description": "If set, the values stored at the given property are used as relationship weights during the computation. If not set, the graph is considered unweighted.",
                },
                "minCommunitySize": {
                    "type": "integer",
                    "description": "Only nodes inside communities larger or equal the given value are returned.",
                },
                "nodeIdentifierProperty": {
                    "type": "string",
                    "description": "The name of a node property to use as node identifier in the result. If provided, the result will include a 'nodeName' column with values from this property.",
                },
                "nodeLabels": {
                    "type": "array",
                    "items": {"type": "string"},
                    "description": "The node labels used to project and run Max-K-Cut  on. Nodes with different node labels will be ignored. Do not specify to run for all nodes",
                },
            },
        },
    ),
    types.Tool(
        name="speaker_listener_label_propagation",
        description="The Speaker-Listener Label Propagation Algorithm (SLLPA) is a variation of the Label Propagation algorithm that is able to detect multiple communities per node. "
        "The GDS implementation is based on the SLPA: Uncovering Overlapping Communities in Social Networks via A Speaker-listener Interaction Dynamic Process publication by Xie et al. "
        "The algorithm is randomized in nature and will not produce deterministic results. "
        "To accommodate this, we recommend using a higher number of iterations.",
        inputSchema={
            "type": "object",
            "properties": {
                "maxIterations": {
                    "type": "integer",
                    "description": "Maximum number of iterations to run.",
                },
                "minAssociationStrength": {
                    "type": "number",
                    "description": "Minimum influence required for a community to retain a node.",
                },
                "partitioning": {
                    "type": "string",
                    "enum": ["AUTO", "RANGE", "DEGREE"],
                    "description": "The partitioning scheme used to divide the work between threads. Available options are AUTO, RANGE, DEGREE.",
                },
                "nodeIdentifierProperty": {
                    "type": "string",
                    "description": "The name of a node property to use as node identifier in the result. If provided, the result will include a 'nodeName' column with values from this property.",
                },
                "nodeLabels": {
                    "type": "array",
                    "items": {"type": "string"},
                    "description": "The node labels used to project and run SLLPA on. Nodes with different node labels will be ignored. Do not specify to run for all nodes",
                },
            },
        },
    ),
]



from typing import Dict, Type
from graphdatascience import GraphDataScience
from .algorithm_handler import AlgorithmHandler
from .centrality_algorithm_handlers import (
    PageRankHandler,
    ArticleRankHandler,
    DegreeCentralityHandler,
    ArticulationPointsHandler,
    BetweennessCentralityHandler,
    BridgesHandler,
    CELFHandler,
    ClosenessCentralityHandler,
    EigenvectorCentralityHandler,
    HarmonicCentralityHandler,
    HITSHandler,
)
from .community_algorithm_handlers import (
    ConductanceHandler,
    HDBSCANHandler,
    KCoreDecompositionHandler,
    K1ColoringHandler,
    KMeansClusteringHandler,
    LabelPropagationHandler,
    LeidenHandler,
    LocalClusteringCoefficientHandler,
    LouvainHandler,
    ModularityMetricHandler,
    ModularityOptimizationHandler,
    StronglyConnectedComponentsHandler,
    TriangleCountHandler,
    WeaklyConnectedComponentsHandler,
    ApproximateMaximumKCutHandler,
    SpeakerListenerLabelPropagationHandler,
)
from .similarity_algorithm_handlers import (
    NodeSimilarityHandler,
    KNearestNeighborsHandler,
)
from .path_algorithm_handlers import (
    DijkstraShortestPathHandler,
    DeltaSteppingShortestPathHandler,
    DijkstraSingleSourceShortestPathHandler,
    AStarShortestPathHandler,
    YensShortestPathsHandler,
    MinimumWeightSpanningTreeHandler,
    MinimumDirectedSteinerTreeHandler,
    PrizeCollectingSteinerTreeHandler,
    AllPairsShortestPathsHandler,
    RandomWalkHandler,
    BreadthFirstSearchHandler,
    DepthFirstSearchHandler,
    BellmanFordSingleSourceShortestPathHandler,
    LongestPathHandler,
)


class AlgorithmRegistry:
    _handlers: Dict[str, Type[AlgorithmHandler]] = {
        # Centrality algorithms
        "article_rank": ArticleRankHandler,
        "articulation_points": ArticulationPointsHandler,
        "betweenness_centrality": BetweennessCentralityHandler,
        "bridges": BridgesHandler,
        "CELF": CELFHandler,
        "closeness_centrality": ClosenessCentralityHandler,
        "degree_centrality": DegreeCentralityHandler,
        "eigenvector_centrality": EigenvectorCentralityHandler,
        "pagerank": PageRankHandler,
        "harmonic_centrality": HarmonicCentralityHandler,
        "HITS": HITSHandler,
        # Community detection algorithms
        "conductance": ConductanceHandler,
        "hdbscan": HDBSCANHandler,
        "k_core_decomposition": KCoreDecompositionHandler,
        "k_1_coloring": K1ColoringHandler,
        "k_means_clustering": KMeansClusteringHandler,
        "label_propagation": LabelPropagationHandler,
        "leiden": LeidenHandler,
        "local_clustering_coefficient": LocalClusteringCoefficientHandler,
        "louvain": LouvainHandler,
        "modularity_metric": ModularityMetricHandler,
        "modularity_optimization": ModularityOptimizationHandler,
        "strongly_connected_components": StronglyConnectedComponentsHandler,
        "triangle_count": TriangleCountHandler,
        "weakly_connected_components": WeaklyConnectedComponentsHandler,
        "approximate_maximum_k_cut": ApproximateMaximumKCutHandler,
        "speaker_listener_label_propagation": SpeakerListenerLabelPropagationHandler,
        # Similarity algorithms
        "node_similarity": NodeSimilarityHandler,
        "k_nearest_neighbors": KNearestNeighborsHandler,
        # Path finding algorithms
        "find_shortest_path": DijkstraShortestPathHandler,
        "delta_stepping_shortest_path": DeltaSteppingShortestPathHandler,
        "dijkstra_single_source_shortest_path": DijkstraSingleSourceShortestPathHandler,
        "a_star_shortest_path": AStarShortestPathHandler,
        "yens_shortest_paths": YensShortestPathsHandler,
        "minimum_weight_spanning_tree": MinimumWeightSpanningTreeHandler,
        "minimum_directed_steiner_tree": MinimumDirectedSteinerTreeHandler,
        "prize_collecting_steiner_tree": PrizeCollectingSteinerTreeHandler,
        "all_pairs_shortest_paths": AllPairsShortestPathsHandler,
        "random_walk": RandomWalkHandler,
        "breadth_first_search": BreadthFirstSearchHandler,
        "depth_first_search": DepthFirstSearchHandler,
        "bellman_ford_single_source_shortest_path": BellmanFordSingleSourceShortestPathHandler,
        "longest_path": LongestPathHandler,
    }

    @classmethod
    def get_handler(cls, name: str, gds: GraphDataScience) -> AlgorithmHandler:
        handler_class = cls._handlers.get(name)
        if handler_class is None:
            raise ValueError(f"Unknown tool: {name}.")
        return handler_class(gds)



import logging
from typing import Any, Dict

from .algorithm_handler import AlgorithmHandler
from .gds import projected_graph
from .node_translator import (
    filter_identifiers,
    translate_ids_to_identifiers,
    translate_identifiers_to_ids,
)

logger = logging.getLogger("mcp_server_neo4j_gds")


class ArticleRankHandler(AlgorithmHandler):
    def article_rank(self, **kwargs):
        with projected_graph(self.gds) as G:
            # If any optional parameter is not None, use that parameter
            args = locals()
            params = {
                k: v
                for k, v in kwargs.items()
                if v is not None
                and k not in ["nodes", "nodeIdentifierProperty", "sourceNodes"]
            }
            node_names = kwargs.get("nodes", None)
            node_identifier_property = kwargs.get("nodeIdentifierProperty")
            source_nodes = kwargs.get("sourceNodes", None)

            # Handle sourceNodes - convert names to IDs if nodeIdentifierProperty is provided
            translate_identifiers_to_ids(
                self.gds, source_nodes, "sourceNodes", node_identifier_property, params
            )

            logger.info(f"ArticleRank parameters: {params}")
            article_ranks = self.gds.articleRank.stream(G, **params)

        translate_ids_to_identifiers(self.gds, node_identifier_property, article_ranks)

        logger.info(f"Filtering ArticleRank results for nodes: {node_names}")
        article_ranks = filter_identifiers(
            self.gds, node_identifier_property, node_names, article_ranks
        )

        return article_ranks

    def execute(self, arguments: Dict[str, Any]) -> Any:
        return self.article_rank(
            nodes=arguments.get("nodes"),
            nodeIdentifierProperty=arguments.get("nodeIdentifierProperty"),
            sourceNodes=arguments.get("sourceNodes"),
            scaler=arguments.get("scaler"),
            dampingFactor=arguments.get("dampingFactor"),
            maxIterations=arguments.get("maxIterations"),
            tolerance=arguments.get("tolerance"),
        )


class ArticulationPointsHandler(AlgorithmHandler):
    def articulation_points(self, **kwargs):
        with projected_graph(self.gds, undirected=True) as G:
            articulation_points = self.gds.articulationPoints.stream(G)

        # Add node names to the results if nodeIdentifierProperty is provided
        node_identifier_property = kwargs.get("nodeIdentifierProperty")
        translate_ids_to_identifiers(
            self.gds, node_identifier_property, articulation_points
        )
        return articulation_points

    def execute(self, arguments: Dict[str, Any]) -> Any:
        return self.articulation_points(
            nodeIdentifierProperty=arguments.get("nodeIdentifierProperty")
        )


class BetweennessCentralityHandler(AlgorithmHandler):
    def betweenness_centrality(self, **kwargs):
        with projected_graph(self.gds) as G:
            params = {
                k: v
                for k, v in kwargs.items()
                if v is not None and k not in ["nodes", "nodeIdentifierProperty"]
            }
            logger.info(f"Betweenness centrality parameters: {params}")
            centrality = self.gds.betweenness.stream(G, **params)

        # Add node names to the results if nodeIdentifierProperty is provided
        node_identifier_property = kwargs.get("nodeIdentifierProperty")
        translate_ids_to_identifiers(self.gds, node_identifier_property, centrality)

        # Filter results by node names if provided
        node_names = kwargs.get("nodes", None)
        centrality = filter_identifiers(
            self.gds, node_identifier_property, node_names, centrality
        )

        return centrality

    def execute(self, arguments: Dict[str, Any]) -> Any:
        return self.betweenness_centrality(
            nodes=arguments.get("nodes"),
            nodeIdentifierProperty=arguments.get("nodeIdentifierProperty"),
            samplingSize=arguments.get("samplingSize"),
            relationshipWeightProperty=arguments.get("relationshipWeightProperty"),
        )


class BridgesHandler(AlgorithmHandler):
    def bridges(self, **kwargs):
        with projected_graph(self.gds, undirected=True) as G:
            bridges_result = self.gds.bridges.stream(G)

        # Add node names to the results if nodeIdentifierProperty is provided
        node_identifier_property = kwargs.get("nodeIdentifierProperty")
        translate_ids_to_identifiers(
            self.gds, node_identifier_property, bridges_result, "from", "fromName"
        )
        translate_ids_to_identifiers(
            self.gds, node_identifier_property, bridges_result, "to", "toName"
        )

        return bridges_result

    def execute(self, arguments: Dict[str, Any]) -> Any:
        return self.bridges(
            nodeIdentifierProperty=arguments.get("nodeIdentifierProperty")
        )


class CELFHandler(AlgorithmHandler):
    def celf(self, **kwargs):
        with projected_graph(self.gds) as G:
            params = {
                k: v
                for k, v in kwargs.items()
                if v is not None and k not in ["nodeIdentifierProperty"]
            }
            logger.info(f"CELF parameters: {params}")
            result = self.gds.influenceMaximization.celf.stream(G, **params)

        # Add node names to the results if nodeIdentifierProperty is provided
        node_identifier_property = kwargs.get("nodeIdentifierProperty")
        translate_ids_to_identifiers(self.gds, node_identifier_property, result)

        return result

    def execute(self, arguments: Dict[str, Any]) -> Any:
        return self.celf(
            seedSetSize=arguments.get("seedSetSize"),
            monteCarloSimulations=arguments.get("monteCarloSimulations"),
            propagationProbability=arguments.get("propagationProbability"),
            nodeIdentifierProperty=arguments.get("nodeIdentifierProperty"),
        )


class ClosenessCentralityHandler(AlgorithmHandler):
    def closeness_centrality(self, **kwargs):
        with projected_graph(self.gds) as G:
            params = {
                k: v
                for k, v in kwargs.items()
                if v is not None and k not in ["nodes", "nodeIdentifierProperty"]
            }
            logger.info(f"Closeness centrality parameters: {params}")
            centrality = self.gds.closeness.stream(G, **params)

        # Add node names to the results if nodeIdentifierProperty is provided
        node_identifier_property = kwargs.get("nodeIdentifierProperty")
        translate_ids_to_identifiers(self.gds, node_identifier_property, centrality)

        # Filter results by node names if provided
        node_names = kwargs.get("nodes", None)
        centrality = filter_identifiers(
            self.gds, node_identifier_property, node_names, centrality
        )

        return centrality

    def execute(self, arguments: Dict[str, Any]) -> Any:
        return self.closeness_centrality(
            nodes=arguments.get("nodes"),
            nodeIdentifierProperty=arguments.get("nodeIdentifierProperty"),
            useWassermanFaust=arguments.get("useWassermanFaust"),
        )


class DegreeCentralityHandler(AlgorithmHandler):
    def degree_centrality(self, **kwargs):
        with projected_graph(self.gds) as G:
            params = {
                k: v
                for k, v in kwargs.items()
                if v is not None and k not in ["nodes", "nodeIdentifierProperty"]
            }
            logger.info(f"Degree centrality parameters: {params}")
            centrality = self.gds.degree.stream(G, **params)

        # Add node names to the results if nodeIdentifierProperty is provided
        node_identifier_property = kwargs.get("nodeIdentifierProperty")
        translate_ids_to_identifiers(self.gds, node_identifier_property, centrality)

        # Filter results by node names if provided
        node_names = kwargs.get("nodes", None)
        centrality = filter_identifiers(
            self.gds, node_identifier_property, node_names, centrality
        )

        return centrality

    def execute(self, arguments: Dict[str, Any]) -> Any:
        return self.degree_centrality(
            nodes=arguments.get("nodes"),
            nodeIdentifierProperty=arguments.get("nodeIdentifierProperty"),
            orientation=arguments.get("orientation"),
            relationshipWeightProperty=arguments.get("relationshipWeightProperty"),
        )


class EigenvectorCentralityHandler(AlgorithmHandler):
    def eigenvector_centrality(self, **kwargs):
        with projected_graph(self.gds) as G:
            params = {
                k: v
                for k, v in kwargs.items()
                if v is not None
                and k not in ["nodes", "nodeIdentifierProperty", "sourceNodes"]
            }
            node_identifier_property = kwargs.get("nodeIdentifierProperty")
            source_nodes = kwargs.get("sourceNodes", None)

            # Handle sourceNodes - convert names to IDs if nodeIdentifierProperty is provided
            translate_identifiers_to_ids(
                self.gds, source_nodes, "sourceNodes", node_identifier_property, params
            )

            logger.info(f"Eigenvector centrality parameters: {params}")
            centrality = self.gds.eigenvector.stream(G, **params)

        # Add node names to the results if nodeIdentifierProperty is provided
        node_identifier_property = kwargs.get("nodeIdentifierProperty")
        translate_ids_to_identifiers(self.gds, node_identifier_property, centrality)

        # Filter results by node names if provided
        node_names = kwargs.get("nodes", None)
        centrality = filter_identifiers(
            self.gds, node_identifier_property, node_names, centrality
        )

        return centrality

    def execute(self, arguments: Dict[str, Any]) -> Any:
        return self.eigenvector_centrality(
            nodes=arguments.get("nodes"),
            nodeIdentifierProperty=arguments.get("nodeIdentifierProperty"),
            maxIterations=arguments.get("maxIterations"),
            tolerance=arguments.get("tolerance"),
            relationshipWeightProperty=arguments.get("relationshipWeightProperty"),
            sourceNodes=arguments.get("sourceNodes"),
            scaler=arguments.get("scaler"),
        )


class PageRankHandler(AlgorithmHandler):
    def pagerank(self, **kwargs):
        with projected_graph(self.gds) as G:
            params = {
                k: v
                for k, v in kwargs.items()
                if v is not None
                and k not in ["nodes", "nodeIdentifierProperty", "sourceNodes"]
            }
            node_identifier_property = kwargs.get("nodeIdentifierProperty")
            source_nodes = kwargs.get("sourceNodes", None)

            # Handle sourceNodes - convert names to IDs if nodeIdentifierProperty is provided
            translate_identifiers_to_ids(
                self.gds, source_nodes, "sourceNodes", node_identifier_property, params
            )
            logger.info(f"Pagerank parameters: {params}")
            pageranks = self.gds.pageRank.stream(G, **params)

        # Add node names to the results if nodeIdentifierProperty is provided
        node_identifier_property = kwargs.get("nodeIdentifierProperty")
        translate_ids_to_identifiers(self.gds, node_identifier_property, pageranks)

        # Filter results by node names if provided
        node_names = kwargs.get("nodes", None)
        pageranks = filter_identifiers(
            self.gds, node_identifier_property, node_names, pageranks
        )

        return pageranks

    def execute(self, arguments: Dict[str, Any]) -> Any:
        return self.pagerank(
            nodes=arguments.get("nodes"),
            nodeIdentifierProperty=arguments.get("nodeIdentifierProperty"),
            sourceNodes=arguments.get("sourceNodes"),
            dampingFactor=arguments.get("dampingFactor"),
            maxIterations=arguments.get("maxIterations"),
            tolerance=arguments.get("tolerance"),
        )


class HarmonicCentralityHandler(AlgorithmHandler):
    def harmonic_centrality(self, **kwargs):
        with projected_graph(self.gds) as G:
            centrality = self.gds.closeness.harmonic.stream(G)

        # Add node names to the results if nodeIdentifierProperty is provided
        node_identifier_property = kwargs.get("nodeIdentifierProperty")
        translate_ids_to_identifiers(self.gds, node_identifier_property, centrality)

        # Filter results by node names if provided
        node_names = kwargs.get("nodes", None)
        centrality = filter_identifiers(
            self.gds, node_identifier_property, node_names, centrality
        )
        return centrality

    def execute(self, arguments: Dict[str, Any]) -> Any:
        return self.harmonic_centrality(
            nodes=arguments.get("nodes"),
            nodeIdentifierProperty=arguments.get("nodeIdentifierProperty"),
        )


class HITSHandler(AlgorithmHandler):
    def hits(self, **kwargs):
        with projected_graph(self.gds) as G:
            params = {
                k: v
                for k, v in kwargs.items()
                if v is not None and k not in ["nodes", "nodeIdentifierProperty"]
            }
            logger.info(f"HITS parameters: {params}")
            result = self.gds.hits.stream(G, **params)

        # Add node names to the results if nodeIdentifierProperty is provided
        node_identifier_property = kwargs.get("nodeIdentifierProperty")
        translate_ids_to_identifiers(self.gds, node_identifier_property, result)

        # Filter results by node names if provided
        node_names = kwargs.get("nodes", None)
        result = filter_identifiers(
            self.gds, node_identifier_property, node_names, result
        )
        return result

    def execute(self, arguments: Dict[str, Any]) -> Any:
        return self.hits(
            nodes=arguments.get("nodes"),
            nodeIdentifierProperty=arguments.get("nodeIdentifierProperty"),
            hitsIterations=arguments.get("hitsIterations"),
            authProperty=arguments.get("authProperty"),
            hubProperty=arguments.get("hubProperty"),
            partitioning=arguments.get("partitioning"),
        )



import logging
from typing import Dict, Any
from .node_translator import (
    filter_identifiers,
    translate_ids_to_identifiers,
)


from .algorithm_handler import AlgorithmHandler
from .gds import projected_graph

logger = logging.getLogger("mcp_server_neo4j_gds")


class ConductanceHandler(AlgorithmHandler):
    def conductance(self, **kwargs):
        with projected_graph(self.gds) as G:
            logger.info(f"Conductance parameters: {kwargs}")
            conductance = self.gds.conductance.stream(G, **kwargs)

        return conductance

    def execute(self, arguments: Dict[str, Any]) -> Any:
        return self.conductance(
            communityProperty=arguments.get("communityProperty"),
            relationshipWeightProperty=arguments.get("relationshipWeightProperty"),
        )


class HDBSCANHandler(AlgorithmHandler):
    def hdbscan(self, **kwargs):
        with projected_graph(self.gds) as G:
            params = {
                k: v
                for k, v in kwargs.items()
                if v is not None and k not in ["nodeIdentifierProperty"]
            }
            logger.info(f"HDBSCAN parameters: {params}")
            hdbscan_result = self.gds.hdbscan.stream(G, **params)

        # Add node names to the results if nodeIdentifierProperty is provided
        node_identifier_property = kwargs.get("nodeIdentifierProperty")
        translate_ids_to_identifiers(self.gds, node_identifier_property, hdbscan_result)

        return hdbscan_result

    def execute(self, arguments: Dict[str, Any]) -> Any:
        return self.hdbscan(
            nodeProperty=arguments.get("nodeProperty"),
            nodeIdentifierProperty=arguments.get("nodeIdentifierProperty"),
            minClusterSize=arguments.get("minClusterSize"),
            samples=arguments.get("samples"),
            leafSize=arguments.get("leafSize"),
        )


class KCoreDecompositionHandler(AlgorithmHandler):
    def k_core_decomposition(self, **kwargs):
        with projected_graph(self.gds, undirected=True) as G:
            logger.info("Running K-Core Decomposition")
            k_core_decomposition_result = self.gds.kcore.stream(G)

        # Add node names to the results if nodeIdentifierProperty is provided
        node_identifier_property = kwargs.get("nodeIdentifierProperty")
        translate_ids_to_identifiers(
            self.gds, node_identifier_property, k_core_decomposition_result
        )
        return k_core_decomposition_result

    def execute(self, arguments: Dict[str, Any]) -> Any:
        return self.k_core_decomposition(
            nodeIdentifierProperty=arguments.get("nodeIdentifierProperty")
        )


class K1ColoringHandler(AlgorithmHandler):
    def k_1_coloring(self, **kwargs):
        with projected_graph(self.gds) as G:
            params = {
                k: v
                for k, v in kwargs.items()
                if v is not None and k not in ["nodeIdentifierProperty"]
            }
            logger.info(f"K-1 Coloring parameters: {params}")
            k1_coloring_result = self.gds.k1coloring.stream(G, **params)

        # Add node names to the results if nodeIdentifierProperty is provided
        node_identifier_property = kwargs.get("nodeIdentifierProperty")
        translate_ids_to_identifiers(
            self.gds, node_identifier_property, k1_coloring_result
        )

        return k1_coloring_result

    def execute(self, arguments: Dict[str, Any]) -> Any:
        return self.k_1_coloring(
            nodeIdentifierProperty=arguments.get("nodeIdentifierProperty"),
            maxIterations=arguments.get("maxIterations"),
            minCommunitySize=arguments.get("minCommunitySize"),
        )


class KMeansClusteringHandler(AlgorithmHandler):
    def k_means_clustering(self, **kwargs):
        with projected_graph(self.gds) as G:
            params = {
                k: v
                for k, v in kwargs.items()
                if v is not None and k not in ["nodeIdentifierProperty"]
            }
            logger.info(f"K-Means Clustering parameters: {params}")
            kmeans_clustering_result = self.gds.kmeans.stream(G, **params)

        # Add node names to the results if nodeIdentifierProperty is provided
        node_identifier_property = kwargs.get("nodeIdentifierProperty")
        translate_ids_to_identifiers(
            self.gds, node_identifier_property, kmeans_clustering_result
        )

        return kmeans_clustering_result

    def execute(self, arguments: Dict[str, Any]) -> Any:
        return self.k_means_clustering(
            nodeProperty=arguments.get("nodeProperty"),
            nodeIdentifierProperty=arguments.get("nodeIdentifierProperty"),
            k=arguments.get("k"),
            maxIterations=arguments.get("maxIterations"),
            deltaThreshold=arguments.get("deltaThreshold"),
            numberOfRestarts=arguments.get("numberOfRestarts"),
            initialSampler=arguments.get("initialSampler"),
            seedCentroids=arguments.get("seedCentroids"),
            computeSilhouette=arguments.get("computeSilhouette"),
        )


class LabelPropagationHandler(AlgorithmHandler):
    def label_propagation(self, **kwargs):
        # Filter out nodeIdentifierProperty as it's not a GDS algorithm parameter
        gds_kwargs = {
            k: v for k, v in kwargs.items() if k not in ["nodeIdentifierProperty"]
        }

        with projected_graph(self.gds) as G:
            logger.info(f"Label Propagation parameters: {gds_kwargs}")
            label_propagation_result = self.gds.labelPropagation.stream(G, **gds_kwargs)

        # Add node names to the results if nodeIdentifierProperty is provided
        node_identifier_property = kwargs.get("nodeIdentifierProperty")
        translate_ids_to_identifiers(
            self.gds, node_identifier_property, label_propagation_result
        )

        return label_propagation_result

    def execute(self, arguments: Dict[str, Any]) -> Any:
        return self.label_propagation(
            maxIterations=arguments.get("maxIterations"),
            nodeWeightProperty=arguments.get("nodeWeightProperty"),
            relationshipWeightProperty=arguments.get("relationshipWeightProperty"),
            seedProperty=arguments.get("seedProperty"),
            consecutiveIds=arguments.get("consecutiveIds"),
            minCommunitySize=arguments.get("minCommunitySize"),
            nodeIdentifierProperty=arguments.get("nodeIdentifierProperty"),
        )


class LeidenHandler(AlgorithmHandler):
    def leiden(self, **kwargs):
        # Filter out nodeIdentifierProperty as it's not a GDS algorithm parameter
        gds_kwargs = {
            k: v for k, v in kwargs.items() if k not in ["nodeIdentifierProperty"]
        }

        with projected_graph(self.gds, undirected=True) as G:
            logger.info(f"Leiden parameters: {gds_kwargs}")
            leiden_result = self.gds.leiden.stream(G, **gds_kwargs)

        # Add node names to the results if nodeIdentifierProperty is provided
        node_identifier_property = kwargs.get("nodeIdentifierProperty")
        translate_ids_to_identifiers(self.gds, node_identifier_property, leiden_result)

        return leiden_result

    def execute(self, arguments: Dict[str, Any]) -> Any:
        return self.leiden(
            maxLevels=arguments.get("maxLevels"),
            gamma=arguments.get("gamma"),
            theta=arguments.get("theta"),
            tolerance=arguments.get("tolerance"),
            includeIntermediateCommunities=arguments.get(
                "includeIntermediateCommunities"
            ),
            seedProperty=arguments.get("seedProperty"),
            minCommunitySize=arguments.get("minCommunitySize"),
            nodeIdentifierProperty=arguments.get("nodeIdentifierProperty"),
        )


class LocalClusteringCoefficientHandler(AlgorithmHandler):
    def local_clustering_coefficient(self, **kwargs):
        # Filter out non-GDS algorithm parameters
        gds_kwargs = {
            k: v
            for k, v in kwargs.items()
            if k not in ["nodeIdentifierProperty", "nodes"]
        }

        with projected_graph(self.gds, undirected=True) as G:
            logger.info(f"Local Clustering Coefficient parameters: {gds_kwargs}")
            local_clustering_coefficient_result = (
                self.gds.localClusteringCoefficient.stream(G, **gds_kwargs)
            )

        # Get filtering parameters
        node_names = kwargs.get("nodes", None)
        node_identifier_property = kwargs.get("nodeIdentifierProperty")

        # Add node names to the results if nodeIdentifierProperty is provided
        translate_ids_to_identifiers(
            self.gds, node_identifier_property, local_clustering_coefficient_result
        )

        # Filter results if nodes parameter provided
        local_clustering_coefficient_result = filter_identifiers(
            self.gds,
            node_identifier_property,
            node_names,
            local_clustering_coefficient_result,
        )

        return local_clustering_coefficient_result

    def execute(self, arguments: Dict[str, Any]) -> Any:
        return self.local_clustering_coefficient(
            triangleCountProperty=arguments.get("triangleCountProperty"),
            nodeIdentifierProperty=arguments.get("nodeIdentifierProperty"),
            nodes=arguments.get("nodes"),
        )


class LouvainHandler(AlgorithmHandler):
    def louvain(self, **kwargs):
        # Filter out nodeIdentifierProperty as it's not a GDS algorithm parameter
        gds_kwargs = {
            k: v for k, v in kwargs.items() if k not in ["nodeIdentifierProperty"]
        }

        with projected_graph(self.gds) as G:
            logger.info(f"Louvain parameters: {gds_kwargs}")
            louvain_result = self.gds.louvain.stream(G, **gds_kwargs)

        # Add node names to the results if nodeIdentifierProperty is provided
        node_identifier_property = kwargs.get("nodeIdentifierProperty")
        translate_ids_to_identifiers(self.gds, node_identifier_property, louvain_result)

        return louvain_result

    def execute(self, arguments: Dict[str, Any]) -> Any:
        return self.louvain(
            relationshipWeightProperty=arguments.get("relationshipWeightProperty"),
            seedProperty=arguments.get("seedProperty"),
            maxLevels=arguments.get("maxLevels"),
            maxIterations=arguments.get("maxIterations"),
            tolerance=arguments.get("tolerance"),
            includeIntermediateCommunities=arguments.get(
                "includeIntermediateCommunities"
            ),
            consecutiveIds=arguments.get("consecutiveIds"),
            minCommunitySize=arguments.get("minCommunitySize"),
            nodeIdentifierProperty=arguments.get("nodeIdentifierProperty"),
        )


class ModularityMetricHandler(AlgorithmHandler):
    def modularity_metric(self, **kwargs):
        with projected_graph(self.gds) as G:
            logger.info(f"Modularity Metric parameters: {kwargs}")
            modularity_metric_result = self.gds.modularity.stream(G, **kwargs)

        return modularity_metric_result

    def execute(self, arguments: Dict[str, Any]) -> Any:
        return self.modularity_metric(
            communityProperty=arguments.get("communityProperty"),
            relationshipWeightProperty=arguments.get("relationshipWeightProperty"),
        )


class ModularityOptimizationHandler(AlgorithmHandler):
    def modularity_optimization(self, **kwargs):
        # Filter out nodeIdentifierProperty as it's not a GDS algorithm parameter
        gds_kwargs = {
            k: v for k, v in kwargs.items() if k not in ["nodeIdentifierProperty"]
        }

        with projected_graph(self.gds) as G:
            logger.info(f"Modularity Optimization parameters: {gds_kwargs}")
            modularity_optimization_result = self.gds.modularityOptimization.stream(
                G, **gds_kwargs
            )

        # Add node names to the results if nodeIdentifierProperty is provided
        node_identifier_property = kwargs.get("nodeIdentifierProperty")
        translate_ids_to_identifiers(
            self.gds, node_identifier_property, modularity_optimization_result
        )

        return modularity_optimization_result

    def execute(self, arguments: Dict[str, Any]) -> Any:
        return self.modularity_optimization(
            maxIterations=arguments.get("maxIterations"),
            tolerance=arguments.get("tolerance"),
            seedProperty=arguments.get("seedProperty"),
            consecutiveIds=arguments.get("consecutiveIds"),
            relationshipWeightProperty=arguments.get("relationshipWeightProperty"),
            minCommunitySize=arguments.get("minCommunitySize"),
            nodeIdentifierProperty=arguments.get("nodeIdentifierProperty"),
        )


class StronglyConnectedComponentsHandler(AlgorithmHandler):
    def strongly_connected_components(self, **kwargs):
        # Filter out nodeIdentifierProperty as it's not a GDS algorithm parameter
        gds_kwargs = {
            k: v for k, v in kwargs.items() if k not in ["nodeIdentifierProperty"]
        }

        with projected_graph(self.gds) as G:
            logger.info(f"Strongly Connected Components parameters: {gds_kwargs}")
            strongly_connected_components_result = self.gds.scc.stream(G, **gds_kwargs)

        # Add node names to the results if nodeIdentifierProperty is provided
        node_identifier_property = kwargs.get("nodeIdentifierProperty")
        translate_ids_to_identifiers(
            self.gds, node_identifier_property, strongly_connected_components_result
        )

        return strongly_connected_components_result

    def execute(self, arguments: Dict[str, Any]) -> Any:
        return self.strongly_connected_components(
            consecutiveIds=arguments.get("consecutiveIds"),
            nodeIdentifierProperty=arguments.get("nodeIdentifierProperty"),
        )


class TriangleCountHandler(AlgorithmHandler):
    def triangle_count(self, **kwargs):
        # Filter out non-GDS algorithm parameters
        gds_kwargs = {
            k: v
            for k, v in kwargs.items()
            if k not in ["nodeIdentifierProperty", "nodes"]
        }

        with projected_graph(self.gds, undirected=True) as G:
            logger.info(f"Triangle Count parameters: {gds_kwargs}")
            triangle_count_result = self.gds.triangleCount.stream(G, **gds_kwargs)

        # Get filtering parameters
        node_names = kwargs.get("nodes", None)
        node_identifier_property = kwargs.get("nodeIdentifierProperty")

        # Add node names to the results if nodeIdentifierProperty is provided
        translate_ids_to_identifiers(
            self.gds, node_identifier_property, triangle_count_result
        )

        # Filter results if nodes parameter provided
        triangle_count_result = filter_identifiers(
            self.gds, node_identifier_property, node_names, triangle_count_result
        )

        return triangle_count_result

    def execute(self, arguments: Dict[str, Any]) -> Any:
        return self.triangle_count(
            maxDegree=arguments.get("maxDegree"),
            nodeIdentifierProperty=arguments.get("nodeIdentifierProperty"),
            nodes=arguments.get("nodes"),
        )


class WeaklyConnectedComponentsHandler(AlgorithmHandler):
    def weakly_connected_components(self, **kwargs):
        # Filter out nodeIdentifierProperty as it's not a GDS algorithm parameter
        gds_kwargs = {
            k: v for k, v in kwargs.items() if k not in ["nodeIdentifierProperty"]
        }

        with projected_graph(self.gds) as G:
            logger.info(f"Weakly Connected Components parameters: {gds_kwargs}")
            weakly_connected_components_result = self.gds.wcc.stream(G, **gds_kwargs)

        # Add node names to the results if nodeIdentifierProperty is provided
        node_identifier_property = kwargs.get("nodeIdentifierProperty")
        translate_ids_to_identifiers(
            self.gds, node_identifier_property, weakly_connected_components_result
        )

        return weakly_connected_components_result

    def execute(self, arguments: Dict[str, Any]) -> Any:
        return self.weakly_connected_components(
            relationshipWeightProperty=arguments.get("relationshipWeightProperty"),
            seedProperty=arguments.get("seedProperty"),
            threshold=arguments.get("threshold"),
            consecutiveIds=arguments.get("consecutiveIds"),
            minComponentSize=arguments.get("minComponentSize"),
            nodeIdentifierProperty=arguments.get("nodeIdentifierProperty"),
        )


class ApproximateMaximumKCutHandler(AlgorithmHandler):
    def approximate_maximum_k_cut(self, **kwargs):
        # Filter out nodeIdentifierProperty as it's not a GDS algorithm parameter
        gds_kwargs = {
            k: v for k, v in kwargs.items() if k not in ["nodeIdentifierProperty"]
        }

        with projected_graph(self.gds) as G:
            logger.info(f"Approximate Maximum K Cut parameters: {gds_kwargs}")
            approximate_maximum_k_cut_result = self.gds.maxkcut.stream(G, **gds_kwargs)

        # Add node names to the results if nodeIdentifierProperty is provided
        node_identifier_property = kwargs.get("nodeIdentifierProperty")
        translate_ids_to_identifiers(
            self.gds, node_identifier_property, approximate_maximum_k_cut_result
        )

        return approximate_maximum_k_cut_result

    def execute(self, arguments: Dict[str, Any]) -> Any:
        return self.approximate_maximum_k_cut(
            k=arguments.get("k"),
            iterations=arguments.get("iterations"),
            vnsMaxNeighborhoodOrder=arguments.get("vnsMaxNeighborhoodOrder"),
            relationshipWeightProperty=arguments.get("relationshipWeightProperty"),
            minCommunitySize=arguments.get("minCommunitySize"),
            nodeIdentifierProperty=arguments.get("nodeIdentifierProperty"),
        )


class SpeakerListenerLabelPropagationHandler(AlgorithmHandler):
    def speaker_listener_label_propagation(self, **kwargs):
        # Filter out nodeIdentifierProperty as it's not a GDS algorithm parameter
        gds_kwargs = {
            k: v for k, v in kwargs.items() if k not in ["nodeIdentifierProperty"]
        }

        with projected_graph(self.gds) as G:
            logger.info(f"Speaker Listener Label Propagation parameters: {gds_kwargs}")
            speaker_listener_label_propagation_result = self.gds.sllpa.stream(
                G, **gds_kwargs
            )

        # Add node names to the results if nodeIdentifierProperty is provided
        node_identifier_property = kwargs.get("nodeIdentifierProperty")
        translate_ids_to_identifiers(
            self.gds,
            node_identifier_property,
            speaker_listener_label_propagation_result,
        )

        return speaker_listener_label_propagation_result

    def execute(self, arguments: Dict[str, Any]) -> Any:
        return self.speaker_listener_label_propagation(
            maxIterations=arguments.get("maxIterations"),
            minAssociationStrength=arguments.get("minAssociationStrength"),
            partitioning=arguments.get("partitioning"),
            nodeIdentifierProperty=arguments.get("nodeIdentifierProperty"),
        )



from graphdatascience import GraphDataScience


def translate_identifiers_to_ids(
    gds: GraphDataScience,
    input_nodes,
    input_nodes_variable_name,
    node_identifier_property,
    call_params,
):
    # Handle input nodes - convert names to IDs if nodeIdentifierProperty is provided
    if input_nodes is not None and node_identifier_property is not None:
        if isinstance(input_nodes, list):
            # Handle list of node names
            query = f"""
                    UNWIND $names AS name
                    MATCH (s)
                    WHERE toLower(s.{node_identifier_property}) = toLower(name)
                    RETURN id(s) as node_id
                    """
            df = gds.run_cypher(
                query,
                params={
                    "names": input_nodes,
                },
            )
            input_node_ids = df["node_id"].tolist()
            call_params[input_nodes_variable_name] = input_node_ids
        else:
            # Handle single  node name
            query = f"""
                    MATCH (s)
                    WHERE toLower(s.{node_identifier_property}) = toLower($name)
                    RETURN id(s) as node_id
                    """
            df = gds.run_cypher(
                query,
                params={
                    "name": input_nodes,
                },
            )
            if not df.empty:
                call_params[input_nodes_variable_name] = int(df["node_id"].iloc[0])
    elif input_nodes is not None:
        # If input_nodes provided but no nodeIdentifierProperty, pass through as-is
        call_params[input_nodes_variable_name] = input_nodes


def translate_ids_to_identifiers(
    gds: GraphDataScience,
    node_identifier_property,
    results,
    id_name="nodeId",
    node_identifier_output_name="nodeName",
):
    if node_identifier_property is not None:
        node_name_values = [
            gds.util.asNode(node_id).get(node_identifier_property)
            for node_id in results[id_name]
        ]
        results[node_identifier_output_name] = node_name_values


def filter_identifiers(
    gds: GraphDataScience,
    node_identifier_property,
    node_names,
    results,
    id_name="nodeId",
):
    if node_names is None:
        return results
    if node_identifier_property is None:
        raise ValueError(
            "If 'nodes' is provided, 'nodeIdentifierProperty' must also be specified."
        )
    query = f"""
            UNWIND $names AS name
            MATCH (s)
            WHERE toLower(s.{node_identifier_property}) = toLower(name)
            RETURN id(s) as node_id
            """
    df = gds.run_cypher(
        query,
        params={
            "names": node_names,
        },
    )
    node_ids = df["node_id"].tolist()
    filtered_results = results[results[id_name].isin(node_ids)]
    return filtered_results



from mcp import types

similarity_tool_definitions = [
    types.Tool(
        name="node_similarity",
        description="The Node Similarity algorithm compares a set of nodes based on the nodes they are connected to. "
        "Two nodes are considered similar if they share many of the same neighbors. "
        "Node Similarity computes pair-wise similarities based on the Jaccard metric, also known as the Jaccard Similarity Score, the Overlap coefficient, also known as the Szymkiewicz–Simpson coefficient, and the Cosine Similarity score. "
        "The first two are most frequently associated with unweighted sets, whereas Cosine with weighted input."
        "Filters on source nodes, target nodes, or both can additionally be provided to compute results for subset of nodes.",
        inputSchema={
            "type": "object",
            "properties": {
                "sourceNodeFilter": {
                    "type": ["integer", "array", "string"],
                    "description": "The source node filter to apply. Accepts a single node id, a List of node ids, or a single label.",
                },
                "targetNodeFilter": {
                    "type": ["integer", "array", "string"],
                    "description": "The target node filter to apply. Accepts a single node id, a List of node ids, or a single label.",
                },
                "similarityCutoff": {
                    "type": "number",
                    "description": "Lower limit for the similarity score to be present in the result. Values must be between 0 and 1.",
                },
                "degreeCutoff": {
                    "type": "integer",
                    "description": "Inclusive lower bound on the node degree for a node to be considered in the comparisons. This value can not be lower than 1.",
                },
                "upperDegreeCutoff": {
                    "type": "integer",
                    "description": "Inclusive upper bound on the node degree for a node to be considered in the comparisons. This value can not be lower than 1.",
                },
                "topK": {
                    "type": "integer",
                    "description": "Limit on the number of scores per node. The K largest results are returned. This value cannot be lower than 1. Use this instead of topN whenever the sourceNode consists of a single node, or it is specifically stated that results are to be computed for each source node",
                },
                "bottomK": {
                    "type": "integer",
                    "description": "Limit on the number of scores per node. The K smallest results are returned. This value cannot be lower than 1.",
                },
                "topN": {
                    "type": "integer",
                    "description": "Global limit on the number of scores computed. The N largest total results are returned. This value cannot be negative, a value of 0 means no global limit.",
                },
                "bottomN": {
                    "type": "integer",
                    "description": "Global limit on the number of scores computed. The N smallest total results are returned. This value cannot be negative, a value of 0 means no global limit.",
                },
                "relationshipWeightProperty": {
                    "type": "string",
                    "description": "Name of the relationship property to use as weights. If unspecified, the algorithm runs unweighted.",
                },
                "similarityMetric": {
                    "type": "string",
                    "enum": ["JACCARD", "OVERLAP", "COSINE"],
                    "description": "The metric used to compute similarity.",
                },
                "useComponents": {
                    "type": "boolean",
                    "description": "If enabled, Node Similarity will use components to improve the performance of the computation, skipping comparisons of nodes in different components. Set to false (Default): the algorithm does not use components, but computes similarity across the entire graph. Set to true: the algorithm uses components, and will compute these components before computing similarity. Set to String: use pre-computed components stored in graph, String is the key for a node property representing components.",
                },
                "nodeIdentifierProperty": {
                    "type": "string",
                    "description": "Property name to use for identifying nodes (e.g., 'name', 'Name', 'title'). Use get_node_properties_keys to find available properties.",
                },
                "nodeLabels": {
                    "type": "array",
                    "items": {"type": "string"},
                    "description": "The node labels used to project and run Node Similarity on. Nodes with different node labels will be ignored. Do not specify to run for all nodes",
                },
            },
        },
    ),
    types.Tool(
        name="k_nearest_neighbors",
        description="The K-Nearest Neighbors algorithm computes a distance value for all node pairs in the graph and creates new relationships between each node and its k nearest neighbors. "
        "The distance is calculated based on node properties. "
        "The input of this algorithm is a homogeneous graph; any node label or relationships type information in the graph is ignored. "
        "The graph does not need to be connected, in fact, existing relationships between nodes will be ignored - apart from random walk sampling if that that initial sampling option is used. "
        "New relationships are created between each node and its k nearest neighbors. "
        "The K-Nearest Neighbors algorithm compares given properties of each node. The k nodes where these properties are most similar are the k-nearest neighbors. "
        "The initial set of neighbors is picked at random and verified and refined in multiple iterations. The number of iterations is limited by the configuration parameter maxIterations. "
        "The algorithm may stop earlier if the neighbor lists only change by a small amount, which can be controlled by the configuration parameter deltaThreshold. "
        "The particular implementation is based on Efficient k-nearest neighbor graph construction for generic similarity measures by Wei Dong et al. "
        "Instead of comparing every node with every other node, the algorithm selects possible neighbors based on the assumption, that the neighbors-of-neighbors of a node are most likely already the nearest one. "
        "The algorithm scales quasi-linear with respect to the node count, instead of being quadratic. "
        "Furthermore, the algorithm only compares a sample of all possible neighbors on each iteration, assuming that eventually all possible neighbors will be seen. "
        "This can be controlled with the configuration parameter sampleRate: A valid sample rate must be in between 0 (exclusive) and 1 (inclusive). "
        "The default value is 0.5. The parameter is used to control the trade-off between accuracy and runtime-performance. "
        "A higher sample rate will increase the accuracy of the result. The algorithm will also require more memory and will take longer to compute. "
        "A lower sample rate will increase the runtime-performance. Some potential nodes may be missed in the comparison and may not be included in the result. "
        "When encountered neighbors have equal similarity to the least similar already known neighbor, randomly selecting which node to keep can reduce the risk of some neighborhoods not being explored. "
        "This behavior is controlled by the configuration parameter perturbationRate. "
        "The output of the algorithm are new relationships between nodes and their k-nearest neighbors. Similarity scores are expressed via relationship properties."
        "Filters on source nodes, target nodes, or both can additionally be provided to compute results for subset of nodes.",
        inputSchema={
            "type": "object",
            "properties": {
                "sourceNodeFilter": {
                    "type": ["integer", "array", "string"],
                    "description": "The source node filter to apply. Accepts a single node id, a List of node ids, or a single label.",
                },
                "targetNodeFilter": {
                    "type": ["integer", "array", "string"],
                    "description": "The target node filter to apply. Accepts a single node id, a List of node ids, or a single label.",
                },
                "nodeProperties": {
                    "type": ["string", "object", "array"],
                    "description": "The node properties to use for similarity computation along with their selected similarity metrics. Accepts a single property key, a Map of property keys to metrics, or a List of property keys and/or Maps, as above.",
                },
                "topK": {
                    "type": "integer",
                    "description": "The number of neighbors to find for each node. The K-nearest neighbors are returned. This value cannot be lower than 1.",
                },
                "sampleRate": {
                    "type": "number",
                    "description": "Sample rate to limit the number of comparisons per node. Value must be between 0 (exclusive) and 1 (inclusive).",
                },
                "deltaThreshold": {
                    "type": "number",
                    "description": "Value as a percentage to determine when to stop early. If fewer updates than the configured value happen, the algorithm stops. Value must be between 0 (exclusive) and 1 (inclusive).",
                },
                "maxIterations": {
                    "type": "integer",
                    "description": "Hard limit to stop the algorithm after that many iterations.",
                },
                "randomJoins": {
                    "type": "integer",
                    "description": "The number of random attempts per node to connect new node neighbors based on random selection, for each iteration.",
                },
                "initialSampler": {
                    "type": "string",
                    "enum": ["uniform", "randomWalk"],
                    "description": 'The method used to sample the first k random neighbors for each node. "uniform" and "randomWalk", both case-insensitive, are valid inputs.',
                },
                "similarityCutoff": {
                    "type": "number",
                    "description": "Filter out from the list of K-nearest neighbors nodes with similarity below this threshold.",
                },
                "perturbationRate": {
                    "type": "number",
                    "description": "The probability of replacing the least similar known neighbor with an encountered neighbor of equal similarity.",
                },
                "seedTargetNodes": {
                    "type": "boolean",
                    "description": "Enable seeding of target nodes. If seeded, every node picks some of the target nodes initially. This guarantees that for every node we can avoid empty result (when the algorithm did not find for it any similar neighbors from the target set). Can only be used if targetNodeFilter is set.",
                },
                "nodeIdentifierProperty": {
                    "type": "string",
                    "description": "Property name to use for identifying nodes (e.g., 'name', 'Name', 'title'). Use get_node_properties_keys to find available properties.",
                },
                "nodeLabels": {
                    "type": "array",
                    "items": {"type": "string"},
                    "description": "The node labels used to project and run k-nearest neighbors on. Nodes with different node labels will be ignored. Do not specify to run for all nodes",
                },
            },
            "required": ["nodeProperties"],
        },
    ),
]



from abc import ABC, abstractmethod
from typing import Dict, Any
from graphdatascience import GraphDataScience


class AlgorithmHandler(ABC):
    def __init__(self, gds: GraphDataScience):
        self.gds = gds

    @abstractmethod
    def execute(self, arguments: Dict[str, Any]) -> Any:
        pass



import pytest


@pytest.mark.asyncio
async def test_node_similarity(mcp_client):
    result = await mcp_client.call_tool(
        "node_similarity", {"nodeIdentifierProperty": "name", "topN": 35}
    )

    assert len(result) == 1

    result_text = result[0]["text"]

    # Verify structure of a result entry
    assert "node1" in result_text
    assert "node2" in result_text
    assert "node1Name" in result_text
    assert "node2Name" in result_text
    assert "similarity" in result_text
    lines = result_text.strip().split("\n")
    data_lines = [line for line in lines[1:] if line.strip()]
    assert len(data_lines) == 35
    assert "Aldgate" in data_lines[34]


@pytest.mark.asyncio
async def test_filtered_node_similarity(mcp_client):
    # test source-filter only
    result = await mcp_client.call_tool(
        "node_similarity",
        {
            "nodeIdentifierProperty": "name",
            "topK": 3,
            "sourceNodeFilter": ["Acton Town"],
        },
    )

    assert len(result) == 1
    result_text = result[0]["text"]
    # Verify structure of a result entry
    assert "node1" in result_text
    assert "node2" in result_text
    assert "node1Name" in result_text
    assert "node2Name" in result_text
    assert "similarity" in result_text
    lines = result_text.strip().split("\n")
    data_lines = [line for line in lines[1:] if line.strip()]
    assert len(data_lines) == 3
    assert "Acton Town" in data_lines[0]

    # test target-filter alone

    result = await mcp_client.call_tool(
        "node_similarity",
        {
            "nodeIdentifierProperty": "name",
            "topK": 3,
            "targetNodeFilter": "Stamford Brook",
        },
    )
    assert len(result) == 1
    result_text = result[0]["text"]
    # Verify structure of a result entry
    assert "node1" in result_text
    assert "node2" in result_text
    assert "node1Name" in result_text
    assert "node2Name" in result_text
    assert "similarity" in result_text
    lines = result_text.strip().split("\n")
    data_lines = [line for line in lines[1:] if line.strip()]
    assert len(data_lines) == 4
    assert "Stamford Brook" in data_lines[0]

    # test combination of filters
    result = await mcp_client.call_tool(
        "node_similarity",
        {
            "nodeIdentifierProperty": "name",
            "topK": 3,
            "sourceNodeFilter": ["Acton Town"],
            "targetNodeFilter": ["Stamford Brook"],
        },
    )

    assert len(result) == 1
    result_text = result[0]["text"]
    # Verify structure of a result entry
    assert "node1" in result_text
    assert "node2" in result_text
    assert "node1Name" in result_text
    assert "node2Name" in result_text
    assert "similarity" in result_text
    lines = result_text.strip().split("\n")
    data_lines = [line for line in lines[1:] if line.strip()]
    assert len(data_lines) == 1
    assert "Acton Town" in data_lines[0]
    assert "Stamford Brook" in data_lines[0]


@pytest.mark.asyncio
async def test_k_nearest_neighbors(mcp_client):
    result = await mcp_client.call_tool(
        "k_nearest_neighbors",
        {"nodeIdentifierProperty": "name", "topK": 3, "nodeProperties": "rail"},
    )

    assert len(result) == 1

    result_text = result[0]["text"]

    # Verify structure of a result entry
    assert "node1" in result_text
    assert "node2" in result_text
    assert "node1Name" in result_text
    assert "node2Name" in result_text
    assert "similarity" in result_text
    lines = result_text.strip().split("\n")
    data_lines = [line for line in lines[1:] if line.strip()]
    assert len(data_lines) == 302 * 3


@pytest.mark.asyncio
async def test_filtered_knn(mcp_client):
    # test source-filter only
    result = await mcp_client.call_tool(
        "k_nearest_neighbors",
        {
            "nodeIdentifierProperty": "name",
            "topK": 3,
            "sourceNodeFilter": ["Acton Town"],
            "nodeProperties": "rail",
        },
    )

    assert len(result) == 1
    result_text = result[0]["text"]
    # Verify structure of a result entry
    assert "node1" in result_text
    assert "node2" in result_text
    assert "node1Name" in result_text
    assert "node2Name" in result_text
    assert "similarity" in result_text
    lines = result_text.strip().split("\n")
    data_lines = [line for line in lines[1:] if line.strip()]
    assert len(data_lines) == 3
    assert "Acton Town" in data_lines[0]

    # test target-filter alone

    result = await mcp_client.call_tool(
        "k_nearest_neighbors",
        {
            "nodeIdentifierProperty": "name",
            "topK": 3,
            "targetNodeFilter": "Stamford Brook",
            "nodeProperties": "rail",
        },
    )
    assert len(result) == 1
    result_text = result[0]["text"]
    # Verify structure of a result entry
    assert "node1" in result_text
    assert "node2" in result_text
    assert "node1Name" in result_text
    assert "node2Name" in result_text
    assert "similarity" in result_text
    lines = result_text.strip().split("\n")
    data_lines = [line for line in lines[1:] if line.strip()]
    assert len(data_lines) > 0
    assert "Stamford Brook" in data_lines[0]

    # test combination of filters
    result = await mcp_client.call_tool(
        "k_nearest_neighbors",
        {
            "nodeIdentifierProperty": "name",
            "topK": 3,
            "sourceNodeFilter": ["Acton Town"],
            "targetNodeFilter": ["Stamford Brook"],
            "seedTargetNodes": True,  # k-nn filtering is a bit special, it might not necessarily find answer if this is not specified (at least for this small example graph)
            "nodeProperties": "rail",
        },
    )

    assert len(result) == 1
    result_text = result[0]["text"]
    # Verify structure of a result entry
    assert "node1" in result_text
    assert "node2" in result_text
    assert "node1Name" in result_text
    assert "node2Name" in result_text
    assert "similarity" in result_text
    lines = result_text.strip().split("\n")
    data_lines = [line for line in lines[1:] if line.strip()]
    assert len(data_lines) == 1
    assert "Acton Town" in data_lines[0]
    assert "Stamford Brook" in data_lines[0]



import logging
from typing import Dict, Any

from .algorithm_handler import AlgorithmHandler
from .gds import projected_graph
from .node_translator import (
    translate_ids_to_identifiers,
    translate_identifiers_to_ids,
)

logger = logging.getLogger("mcp_server_neo4j_gds")


class NodeSimilarityHandler(AlgorithmHandler):
    def node_similarity(self, **kwargs):
        with projected_graph(self.gds) as G:
            params = {
                k: v
                for k, v in kwargs.items()
                if v is not None
                and k
                not in [
                    "nodeIdentifierProperty",
                    "sourceNodeFilter",
                    "targetNodeFilter",
                ]
            }
            node_identifier_property = kwargs.get("nodeIdentifierProperty")
            source_nodes = kwargs.get("sourceNodeFilter", None)
            target_nodes = kwargs.get("targetNodeFilter", None)
            translate_identifiers_to_ids(
                self.gds,
                source_nodes,
                "sourceNodeFilter",
                node_identifier_property,
                params,
            )
            translate_identifiers_to_ids(
                self.gds,
                target_nodes,
                "targetNodeFilter",
                node_identifier_property,
                params,
            )
            logger.info(f"Node Similarity parameters: {params}")
            node_similarity_result = self.gds.nodeSimilarity.filtered.stream(
                G, **params
            )

        # Add node names to the results if nodeIdentifierProperty is provided
        node_identifier_property = kwargs.get("nodeIdentifierProperty")
        translate_ids_to_identifiers(
            self.gds,
            node_identifier_property,
            node_similarity_result,
            "node1",
            "node1Name",
        )
        translate_ids_to_identifiers(
            self.gds,
            node_identifier_property,
            node_similarity_result,
            "node2",
            "node2Name",
        )
        return node_similarity_result

    def execute(self, arguments: Dict[str, Any]) -> Any:
        return self.node_similarity(
            nodeIdentifierProperty=arguments.get("nodeIdentifierProperty"),
            sourceNodeFilter=arguments.get("sourceNodeFilter"),
            targetNodeFilter=arguments.get("targetNodeFilter"),
            similarityCutoff=arguments.get("similarityCutoff"),
            degreeCutoff=arguments.get("degreeCutoff"),
            upperDegreeCutoff=arguments.get("upperDegreeCutoff"),
            topK=arguments.get("topK"),
            bottomK=arguments.get("bottomK"),
            topN=arguments.get("topN"),
            bottomN=arguments.get("bottomN"),
            relationshipWeightProperty=arguments.get("relationshipWeightProperty"),
            similarityMetric=arguments.get("similarityMetric"),
            useComponents=arguments.get("useComponents"),
        )


class KNearestNeighborsHandler(AlgorithmHandler):
    def k_nearest_neighbors(self, **kwargs):
        with projected_graph(self.gds) as G:
            params = {
                k: v
                for k, v in kwargs.items()
                if v is not None
                and k
                not in [
                    "nodeIdentifierProperty",
                    "sourceNodeFilter",
                    "targetNodeFilter",
                ]
            }
            node_identifier_property = kwargs.get("nodeIdentifierProperty")
            source_nodes = kwargs.get("sourceNodeFilter", None)
            target_nodes = kwargs.get("targetNodeFilter", None)
            translate_identifiers_to_ids(
                self.gds,
                source_nodes,
                "sourceNodeFilter",
                node_identifier_property,
                params,
            )
            translate_identifiers_to_ids(
                self.gds,
                target_nodes,
                "targetNodeFilter",
                node_identifier_property,
                params,
            )

            logger.info(f"K-Nearest Neighbors parameters: {kwargs}")
            k_nearest_neighbors_result = self.gds.knn.filtered.stream(G, **params)

        # Add node names to the results if nodeIdentifierProperty is provided
        node_identifier_property = kwargs.get("nodeIdentifierProperty")
        translate_ids_to_identifiers(
            self.gds,
            node_identifier_property,
            k_nearest_neighbors_result,
            "node1",
            "node1Name",
        )
        translate_ids_to_identifiers(
            self.gds,
            node_identifier_property,
            k_nearest_neighbors_result,
            "node2",
            "node2Name",
        )

        return k_nearest_neighbors_result

    def execute(self, arguments: Dict[str, Any]) -> Any:
        return self.k_nearest_neighbors(
            nodeIdentifierProperty=arguments.get("nodeIdentifierProperty"),
            sourceNodeFilter=arguments.get("sourceNodeFilter"),
            targetNodeFilter=arguments.get("targetNodeFilter"),
            nodeProperties=arguments.get("nodeProperties"),
            topK=arguments.get("topK"),
            sampleRate=arguments.get("sampleRate"),
            deltaThreshold=arguments.get("deltaThreshold"),
            maxIterations=arguments.get("maxIterations"),
            randomJoins=arguments.get("randomJoins"),
            initialSampler=arguments.get("initialSampler"),
            similarityCutoff=arguments.get("similarityCutoff"),
            perturbationRate=arguments.get("perturbationRate"),
            seedTargetNodes=arguments.get("seedTargetNodes"),
        )



import logging
import os
import sys
from dotenv import load_dotenv

from . import server
import asyncio
import argparse


logger = logging.getLogger("mcp_server_neo4j_gds")
logger.handlers.clear()
handler = logging.StreamHandler(sys.stderr)
handler.setFormatter(
    logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
)
logger.addHandler(handler)
logger.setLevel(logging.INFO)
logger.propagate = False


def main():
    """Main entry point for the package."""
    load_dotenv("../../../.env")
    parser = argparse.ArgumentParser(description="Neo4j GDS MCP Server")
    parser.add_argument(
        "--db-url", default=os.environ.get("NEO4J_URI"), help="URL to Neo4j database"
    )
    parser.add_argument(
        "--username",
        default=os.environ.get("NEO4J_USERNAME", "neo4j"),
        help="Username for Neo4j database",
    )
    parser.add_argument(
        "--password",
        default=os.environ.get("NEO4J_PASSWORD"),
        help="Password for Neo4j database",
    )
    parser.add_argument(
        "--database",
        default=os.environ.get("NEO4J_DATABASE"),
        help="Database name to connect to (optional). By default, the server will connect to the 'neo4j' database.",
    )

    args = parser.parse_args()

    asyncio.run(
        server.main(
            db_url=args.db_url,
            username=args.username,
            password=args.password,
            database=args.database,
        )
    )


__all__ = ["main", "server"]
